#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <netinet/in.h>
#include <netinet/ip_icmp.h>
#include <linux/if.h>
#include <linux/netlink.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_conntrack_tuple.h>
#include <linux/netfilter_ipv4/ip_conntrack_tcp.h>
#include <linux/netfilter_ipv4/ip_conntrack_icmp.h>
#include <linux/netfilter_ipv4/ip_conntrack_ftp.h>
#include <linux/nfnetlink.h>

#include "utility.h"
#include "display.h"


char *pr_ctinfo(unsigned short info)
{
        static char buff[250];
        int len = 0;

        memset(buff, '\0', sizeof(buff));
        switch(CTINFO2DIR(info)) {
                case (IP_CT_DIR_REPLY):
                        len += sprintf(buff + len, "Reply");
                        break;
                case (IP_CT_DIR_ORIGINAL):
                        len += sprintf(buff + len, "Original");
                        break;
                default:
                        len += sprintf(buff + len, "Unknown %d",
                                CTINFO2DIR(info));
                        break;
        }

        len += sprintf(buff + len, ", ");
        switch(info) {
                case (IP_CT_NEW):
                        len += sprintf(buff + len, "New");
                        break;
                case (IP_CT_ESTABLISHED):
                        len += sprintf(buff + len, "Established");
                        break;
                case (IP_CT_RELATED):
                        len += sprintf(buff + len, "Related");
                        break;
                case (IP_CT_IS_REPLY):
                        len += sprintf(buff + len, "Reply");
                        break;
                case (IP_CT_DELETE):
                        len += sprintf(buff + len, "Delete");
                        break;
                default:
                        len += sprintf(buff + len, "Unknown %d", info);
                        break;
        }

        return (buff);
}

char *display_sockaddr_nl(struct sockaddr_nl *nl)
{
        static char buff[250];

        memset(buff, '\0', sizeof(buff));
        sprintf(buff, "Family (%s) Pid (%d) Groups (%02X)",
                pr_af_family(nl->nl_family), nl->nl_pid, nl->nl_groups);

        return (buff);
}

char *pr_nlmsg_type(unsigned short type)
{
        static char buff[20];

        memset(buff, '\0', sizeof(buff));
        switch(type) {
                case (NLMSG_NOOP):
                        sprintf(buff, "No operation");
                        break;
                case (NLMSG_ERROR):
                        sprintf(buff, "Error");
                        break;
                case (NLMSG_DONE):
                        sprintf(buff, "Done");
                        break;
                case (NLMSG_OVERRUN):
                        sprintf(buff, "Overrun");
                        break;
                default:
                        sprintf(buff, "Unknown (%02X)", type);
                        break;
        }

        return (buff);
}

char *display_nlmsg(struct nlmsghdr *nl)
{
        static char buff[100];

        memset(buff, '\0', sizeof(buff));
        sprintf(buff, "Length (%d) Type (%s) Flags (%02X) "
                "Sequence (%d) Pid (%d)",
                nl->nlmsg_len, pr_nlmsg_type(nl->nlmsg_type), nl->nlmsg_flags,
                nl->nlmsg_seq, nl->nlmsg_pid);

        return (buff);
}

char *display_nlmsg_error(struct nlmsgerr *nlerr, int msg_len)
{
        static char buff[50];
        int len;

        memset(buff, '\0', sizeof(buff));
        len = sprintf(buff, "%s", "payload contains error data: ");
        if(msg_len < NLMSG_LENGTH(sizeof(struct nlmsgerr)))
                len += sprintf(buff + len, "error packet is truncated.");
        else
        {
                errno = -nlerr->error;
                len += sprintf(buff + len, "%s.", strerror(errno));
        }

        return (buff);
}

char *display_tuple_flat(struct ip_conntrack_tuple *tuple)
{
        static char buff[250];
        char psb[20];
        int len = 0;

        memset(buff, '\0', sizeof(buff));
	len += sprintf(buff + len, "%s:", in_ntoa(tuple->src.ip));
        switch(tuple->dst.protonum) {
                case (IPPROTO_ICMP):
                        len += sprintf(buff + len, "Icmp (id %d)",
                                ntohs(tuple->src.u.icmp.id));
                        break;
                case (IPPROTO_TCP):
                        sprintf(psb, "%d", ntohs(tuple->src.u.tcp.port));
                        len += sprintf(buff + len, "%s", psb);
                        break;
                case (IPPROTO_UDP):
                        sprintf(psb, "%d", ntohs(tuple->src.u.udp.port));
                        len += sprintf(buff + len, "%s", psb);
                        break;
                default:
                        len += sprintf(buff + len, "Unknown");
                        break;
        }

	len += sprintf(buff + len, "->");
        len += sprintf(buff + len, "%s:", in_ntoa(tuple->dst.ip));
        switch(tuple->dst.protonum) {
                case (IPPROTO_ICMP):
                        len += sprintf(buff + len, "Icmp (%s, code %d)",
                                pr_icmp_type(tuple->dst.u.icmp.type),
                                tuple->dst.u.icmp.code);
                        break;
                case (IPPROTO_TCP):
                        sprintf(psb, "%d", ntohs(tuple->dst.u.tcp.port));
                        len += sprintf(buff + len, "%s", psb);
                        break;
                case (IPPROTO_UDP):
                        sprintf(psb, "%d", ntohs(tuple->dst.u.udp.port));
                        len += sprintf(buff + len, "%s", psb);
                        break;
                default:
                        len += sprintf(buff + len, "Unknown");
                        break;
        }

        return (buff);
}

char *pr_tcp_state(unsigned char state)
{
        static char buff[25];

        memset(buff, '\0', sizeof(buff));
        switch(state) {
                case (TCP_CONNTRACK_NONE):
                        sprintf(buff, "none");
                        break;
                case (TCP_CONNTRACK_ESTABLISHED):
                        sprintf(buff, "established");
                        break;
                case (TCP_CONNTRACK_SYN_SENT):
                        sprintf(buff, "syn sent");
                        break;
                case (TCP_CONNTRACK_SYN_RECV):
                        sprintf(buff, "syn received");
                        break;
                case (TCP_CONNTRACK_FIN_WAIT):
                        sprintf(buff, "fin wait");
                        break;
                case (TCP_CONNTRACK_TIME_WAIT):
                        sprintf(buff, "time wait");
                        break;
                case (TCP_CONNTRACK_CLOSE):
                        sprintf(buff, "close");
                        break;
                case (TCP_CONNTRACK_CLOSE_WAIT):
                        sprintf(buff, "close wait");
                        break;
                case (TCP_CONNTRACK_LAST_ACK):
                        sprintf(buff, "last ack");
                        break;
                case (TCP_CONNTRACK_LISTEN):
                        sprintf(buff, "listen");
                        break;
                default:
                        sprintf(buff, "unknown");
                        break;
        }

        return (buff);
}

char *pr_icmp_type(unsigned char type)
{
        static char buff[25];

        memset(buff, '\0', sizeof(buff));
        switch(type) {
                case (ICMP_ECHOREPLY):
                        sprintf(buff, "%s", "Echo Reply");
                        break;
                case (ICMP_DEST_UNREACH):
                        sprintf(buff, "%s", "Destination Unreachable");
                        break;
                case (ICMP_SOURCE_QUENCH):
                        sprintf(buff, "%s", "Source Quench");
                        break;
                case (ICMP_REDIRECT):
                        sprintf(buff, "%s", "Redirect");
                        break;
                case (ICMP_ECHO):
                        sprintf(buff, "%s", "Echo Request");
                        break;
                case (ICMP_TIME_EXCEEDED):
                        sprintf(buff, "%s", "Time Exceeded");
                        break;
                case (ICMP_PARAMETERPROB):
                        sprintf(buff, "%s", "Parameter Problem");
                        break;
                case (ICMP_TIMESTAMP):
                        sprintf(buff, "%s", "Timestamp Request");
                        break;
                case (ICMP_TIMESTAMPREPLY):
                        sprintf(buff, "%s", "Timestamp Reply");
                        break;
                case (ICMP_INFO_REQUEST):
                        sprintf(buff, "%s", "Information Request");
                        break;
                case (ICMP_INFO_REPLY):
                        sprintf(buff, "%s", "Information Reply");
                        break;
                case (ICMP_ADDRESS):
                        sprintf(buff, "%s", "Address Mask Request");
                        break;
                case (ICMP_ADDRESSREPLY):
                        sprintf(buff, "%s", "Address Mask Reply");
                        break;
                default:
                        sprintf(buff, "%s", "unknown");
                        break;
        }

        return (buff);
}

char *pr_ct_status(unsigned long status)
{
        static char buf[100];

        if(!(status & IPS_SEEN_REPLY))
                sprintf(buf, "[UNREPLIED] ");
        if(status & IPS_ASSURED)
                sprintf(buf, "[ASSURED] ");
        return (buf);
}

char *pr_man_type(unsigned short m)
{
        static char buf[10];

        switch(m) {
                case (IP_NAT_MANIP_SRC):
                        sprintf(buf, "Src");
                        break;

                case (IP_NAT_MANIP_DST):
                        sprintf(buf, "Dst");
                        break;
        }

        return (buf);
}

char *pr_man_direction(unsigned char d)
{
        static char buf[10];

        switch(d) {
                case (IP_CT_DIR_REPLY):
                        sprintf(buf, "Rply");
                        break;
                case (IP_CT_DIR_ORIGINAL):
                        sprintf(buf, "Orig");
                        break;
        }

        return (buf);
}

char *pr_proto(int p)
{
        static char buff[25];

        memset(buff, '\0', sizeof(buff));
        switch (p) {
                case (0):
                        sprintf(buff, "any");
                        break;
                case (IPPROTO_ICMP):
                        sprintf(buff, "icmp");
                        break;
                case (IPPROTO_IGMP):
                        sprintf(buff, "igmp");
                        break;
                case (IPPROTO_IPIP):
                        sprintf(buff, "ipip");
                        break;
                case (IPPROTO_TCP):
                        sprintf(buff, "tcp");
                        break;
                case (IPPROTO_EGP):
                        sprintf(buff, "egp");
                        break;
                case (IPPROTO_PUP):
                        sprintf(buff, "pup");
                        break;
                case (IPPROTO_UDP):
                        sprintf(buff, "udp");
                        break;
                case (IPPROTO_IDP):
                        sprintf(buff, "idp");
                        break;
                case (IPPROTO_RSVP):
                        sprintf(buff, "rsvp");
                        break;
                case (IPPROTO_GRE):
                        sprintf(buff, "gre");
                        break;
                case (IPPROTO_IPV6):
                        sprintf(buff, "ipv6");
                        break;
                case (IPPROTO_PIM):
                        sprintf(buff, "pim");
                        break;
                case (IPPROTO_ESP):
                        sprintf(buff, "esp");
                        break;
                case (IPPROTO_AH):
                        sprintf(buff, "ah");
                        break;
                case (IPPROTO_COMP):
                        sprintf(buff, "comp");
                        break;
                case (IPPROTO_RAW):
                        sprintf(buff, "raw");
                        break;
                default:
                        sprintf(buff, "unknown");
                        break;
        }

        return (buff);
}

char *pr_af_family(unsigned char family)
{
        static char buff[30];

        memset(buff, '\0', sizeof(buff));
        switch(family) {
                case (AF_LOCAL):
                        sprintf(buff, "AF_LOCAL");
                        break;
                case (AF_INET):
                        sprintf(buff, "IPv4");
                        break;
                case (AF_AX25):
                        sprintf(buff, "AX.25");
                        break;
                case (AF_IPX):
                        sprintf(buff, "IPX");
                        break;
                case (AF_APPLETALK):
                        sprintf(buff, "Appletalk");
                        break;
                case (AF_NETROM):
                        sprintf(buff, "AF_NETROM");
                        break;
                case (AF_BRIDGE):
                        sprintf(buff, "AF_BRIDGE");
                        break;
                case (AF_ATMPVC):
                        sprintf(buff, "AF_ATMPVC");
                        break;
                case (AF_X25):
                        sprintf(buff, "X.25");
                        break;
                case (AF_INET6):
                        sprintf(buff, "IPv6");
                        break;
                case (AF_ROSE):
                        sprintf(buff, "AF_ROSE");
                        break;
                case (AF_DECnet):
                        sprintf(buff, "DECnet");
                        break;
                case (AF_NETBEUI):
                        sprintf(buff, "AF_NETBEUI");
                        break;
                case (AF_SECURITY):
                        sprintf(buff, "AF_SECURITY");
                        break;
                case (AF_KEY):
                        sprintf(buff, "AF_KEY");
                        break;
                case (AF_NETLINK):
                        sprintf(buff, "AF_NETLINK");
                        break;
                case (AF_PACKET):
                        sprintf(buff, "AF_PACKET");
                        break;
                case (AF_ASH):
                        sprintf(buff, "AF_ASH");
                        break;
                case (AF_ECONET):
                        sprintf(buff, "AF_ECONET");
                        break;
                case (AF_ATMSVC):
                        sprintf(buff, "AF_ATMSVC");
                        break;
                case (AF_SNA):
                        sprintf(buff, "AF_SNA");
                        break;
                case (AF_IRDA):
                        sprintf(buff, "AF_IRDA");
                        break;
                case (AF_PPPOX):
                        sprintf(buff, "AF_PPPOX");
                        break;
                default:
                        sprintf(buff, "Unknown (%d)", family);
        }

        return (buff);
}

#define IP_PARTS_NATIVE(n)                      \
(unsigned int)((n)>>24)&0xFF,                   \
(unsigned int)((n)>>16)&0xFF,                   \
(unsigned int)((n)>>8)&0xFF,                    \
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

char *pr_ipt_ip(struct ipt_ip *ip)
{
        static char buff[300];
        int len;

        memset(buff, '\0', sizeof(buff));
        len = sprintf(buff, "Src IP: (%u.%u.%u.%u/%u.%u.%u.%u) "
                "Dst IP: (%u.%u.%u.%u/%u.%u.%u.%u)\n",
                IP_PARTS(ip->src.s_addr), IP_PARTS(ip->smsk.s_addr),
                IP_PARTS(ip->dst.s_addr), IP_PARTS(ip->dmsk.s_addr));
        len += sprintf(buff + len, "In Iface: (%s)/(%s) Out Iface: (%s)/(%s)\n",                ip->iniface, ip->iniface_mask, ip->outiface, ip->outiface_mask);        len += sprintf(buff + len, "Protocol: (%s) Flags: (%02X) "
                "InvFlags: (%02X)", pr_proto(ip->proto),
                ip->flags, ip->invflags);

        return (buff);
}

char *pr_nfcache(unsigned int nfcache)
{
        static char buff[100];
        int len = 0;

        memset(buff, '\0', sizeof(buff));
	if (nfcache == 0)
		len += sprintf(buff + len, "None ");
        if (nfcache & NFC_ALTERED)
                len += sprintf(buff + len, "ALTERED ");
        if (nfcache & NFC_UNKNOWN)
                len += sprintf(buff + len, "UNKNOWN ");
        if (nfcache & NFC_IP_SRC)
                len += sprintf(buff + len, "IP_SRC ");
        if (nfcache & NFC_IP_DST)
                len += sprintf(buff + len, "IP_DST ");
        if (nfcache & NFC_IP_IF_IN)
                len += sprintf(buff + len, "IP_IF_IN ");
        if (nfcache & NFC_IP_IF_OUT)
                len += sprintf(buff + len, "IP_IF_OUT ");
        if (nfcache & NFC_IP_TOS)
                len += sprintf(buff + len, "IP_TOS ");
        if (nfcache & NFC_IP_PROTO)
                len += sprintf(buff + len, "IP_PROTO ");
        if (nfcache & NFC_IP_OPTIONS)
                len += sprintf(buff + len, "IP_OPTIONS ");
        if (nfcache & NFC_IP_TCPFLAGS)
                len += sprintf(buff + len, "IP_TCPFLAGS ");
        if (nfcache & NFC_IP_SRC_PT)
                len += sprintf(buff + len, "IP_SRC_PT ");
        if (nfcache & NFC_IP_DST_PT)
                len += sprintf(buff + len, "IP_DST_PT ");
        if (nfcache & NFC_IP_PROTO_UNKNOWN)
                len += sprintf(buff + len, "IP_PROTO_UNKNOWN ");

        if(len)
                buff[len - 1] = '\0';

        return (buff);
}

char *pr_match(struct ipta_info *m)
{
	static char buff[100];

        sprintf(buff, "Name: (%s) Size: (%d)",
                m->name, m->size);

	return (buff);
}

char *pr_target(struct ipta_info *t)
{
        static char buff[100];
        int len = 0;

        memset(buff, '\0', sizeof(buff));
        len += sprintf(buff + len, "Name: (%s) [%u] ",
                t->name, t->size);
        if (strcmp(t->name, IPT_STANDARD_TARGET) == 0)
        {
                int pos = *(int *)t->data;
                if (pos < 0)
                        len += sprintf(buff + len, "Verdict: (%s)",
                               pos == -NF_ACCEPT-1 ? "NF_ACCEPT"
                               : pos == -NF_DROP-1 ? "NF_DROP"
                               : pos == -NF_QUEUE-1 ? "NF_QUEUE"
                               : pos == IPT_RETURN ? "IPT_RETURN"
                               : "UNKNOWN");
                else
                        len += sprintf(buff + len, "Verdict: (%u)", pos);
        }
        else
	{
                if (strcmp(t->name, IPT_ERROR_TARGET) == 0)
                        len += sprintf(buff + len, "Error: (%s)", t->data);
        }

        return (buff);
}
